"""Time-series embedding functions."""
import typing as t

import numpy as np
import scipy.spatial

try:
    import pymfe._utils as _utils

except ImportError:
    pass

try:
    import pymfe.autocorr as autocorr

except ImportError:
    pass

try:
    import pymfe.info_theory as info_theory

except ImportError:
    pass


def embed_ts(
    ts: np.ndarray, dim: int, lag: int = 1, include_val: bool = True
) -> np.ndarray:
    """Embbed a time-series in dimension ``dim``.

    Arguments
    ---------
    ts : :obj:`np.ndarray`, shape: (ts.size,)
        One-dimensional time-series.

    dim : int
        Dimension of the embedding.

    lag : int, optional (default=1)
        Lag of the time-series.

    include_val : bool, optional (default=False)
        Include the value itself on its own embedding.

    Returns
    -------
    :obj:`np.ndarray`, shape: (ts.size - dim * lag, dim)
        Embbed time-series.
    """
    if dim <= 0:
        raise ValueError("'dim' must be positive (got {}).".format(dim))

    if lag <= 0:
        raise ValueError("'lag' must be positive (got {}).".format(lag))

    if dim * lag > ts.size:
        raise ValueError(
            "'dim * lag' ({}) can't be larger than the "
            "time-series length ({}).".format(dim * lag, ts.size)
        )

    if include_val:
        dim -= 1

    ts_emb = np.zeros(
        (ts.size - dim * lag, dim + int(include_val)), dtype=ts.dtype
    )

    shift_inds = lag * (dim - 1 - np.arange(-int(include_val), dim))

    for i in np.arange(ts_emb.shape[0]):
        ts_emb[i, :] = ts[i + shift_inds]

    return ts_emb


def nn(
    embed: np.ndarray, metric: str = "chebyshev", p: t.Union[int, float] = 2
) -> np.ndarray:
    """Return the Nearest neighbor of each embedded time-series observation."""
    dist_mat = scipy.spatial.distance.cdist(embed, embed, metric=metric, p=p)

    # Note: prevent nearest neighbor be the instance itself, and also
    # be exact equal instances. We follow Cao's recommendation to pick
    # the next nearest neighbor when this happens.
    dist_mat[np.isclose(dist_mat, 0.0)] = np.inf

    nn_inds = np.argmin(dist_mat, axis=1)

    return nn_inds, dist_mat[nn_inds, np.arange(nn_inds.size)]


def embed_dim_fnn(
    ts: np.ndarray,
    lag: int,
    dims: t.Union[int, t.Sequence[int]] = 16,
    rtol: t.Union[int, float] = 10,
    atol: t.Union[int, float] = 2,
    ts_scaled: t.Optional[np.ndarray] = None,
) -> int:
    """Estimation of the False Nearest Neighbors proportion for each dimension.

    The False Nearest Neighbors calculates the average number of false nearest
    neighbors of each time-series observations, given a fixed embedding
    dimension.

    A false nearest neighbors are a pair of instances that are farther apart
    in the appropriate embedding dimension, but close together in a smaller
    dimension simply because both are projected in a innapropriate dimension.
    Sure enough, we could have just use a `sufficiently large` embedding
    dimension to remove all possibility of false nearest neighbors. However,
    this strategy may imply in a lack of computational effciency, and all
    statistical concerns that may arise in high dimensional data analysis.
    The idea behind of analysing the proportion of false neighbors is to
    estimate the minimum embedding dimension that makes only true neighbors
    be close together in that given space.

    Thus, it is expected that, given the appropriate embedding dimension, the
    proportion of false neighbors will be close to zero.

    Differently from the reference paper, here we are using the Chebyshev
    distance (or maximum norm distance) rather than the Euclidean distance.

    Parameters
    ----------
    ts : :obj:`np.ndarray`
        One-dimensional time-series values.

    lag : int
        Embedding lag. You may want to check the `embed_lag` function
        documentation for embedding lag estimation. Must be a stricly
        positive value.

    dims : int or sequence of int
        Dimensions to estimate the Cao's `E1` and `E2` statistic values.
        If integer, estimate all dimensions from 1 up to the given number.
        If a sequence of integers, estimate the FNN proportion for all
        given dimensions, and return the corresponding values in the same
        order of the given dimensions.
        All dimensions with non-positive values will receive a `np.nan`.

    rtol : float, optional (default=10)
        Relative tolerance between the relative difference of the distances
        between each observation and its nearest neighbor $D_{d}$ in a given
        dimension $d$, and the distance $D_{d+1}$ of the observation and the
        same nearest neighbor in the next embedding dimension. It is used in
        the first criteria from the reference paper to define which instances
        are false neighbors. The default value (10) is the recommended value
        from the original paper, and it means that nearest neighbors that are
        ten times farther in the next dimension relative to the distance in
        the current dimension are considered false nearest neighbors.

    atol : float, optional (default=2)
        Number of time-series standard deviations that an observation and
        its nearest neighbor must be in the next dimension in order to be
        considered false neighbors. This is the reference paper's second
        criteria.

    ts_scaled : :obj:`np.ndarray`, optional
        Standardized time-series values. Used to take advantage of
        precomputations.

    Returns
    -------
    :obj:`np.ndarray`
        Proportion of false nearest neighbos for each given dimension. It is
        used the union of both criterium to determine whether a pair of
        neighbors are false neighbors in a fixed embedding dimension (i.e.,
        any pair of neighbors considered false in either of the criterium
        alone are considered false).

    References
    ----------
    .. [1] Determining embedding dimension for phase-space reconstruction using
        a geometrical construction, Kennel, Matthew B. and Brown, Reggie and
        Abarbanel, Henry D. I., Phys. Rev. A, volume 45, 1992, American
        Physical Society.
    """
    if lag <= 0:
        raise ValueError("'lag' must be positive (got {}).".format(lag))

    _dims: t.Sequence[int]

    if np.isscalar(dims):
        _dims = np.arange(1, int(dims) + 1)  # type: ignore

    else:
        _dims = np.asarray(dims, dtype=int)

    ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

    fnn_prop = np.zeros(len(_dims), dtype=float)

    # Note: since we are using the standardized time-series, its standard
    # deviation is always 1. However, we keep this variable to make clear
    # the correspondence between the reference paper's formulas and what
    # are programmed here.
    ts_std = 1.0  # = np.std(ts_scaled)

    for ind, dim in enumerate(_dims):
        try:
            emb_next = embed_ts(ts=ts_scaled, lag=lag, dim=dim + 1)
            emb_cur = emb_next[:, 1:]

        except ValueError:
            fnn_prop[ind] = np.nan
            continue

        nn_inds, dist_cur = nn(embed=emb_cur)

        emb_next_abs_diff = np.abs(emb_next[:, 0] - emb_next[nn_inds, 0])
        dist_next = np.maximum(dist_cur, emb_next_abs_diff)

        # Note: in the reference paper, there were three criteria for
        # determining what is a False Nearest Neighbor. The first and second
        # one are, respectively, related to the `crit_1` and `crit_2`
        # variables. The third criteria is the union of the criteria, which
        # means that the observation is considered a False Neighbor if either
        # criteria accuses it as such. Here, we are using the third and
        # therefore the most conservative criteria.
        crit_1 = emb_next_abs_diff > rtol * dist_cur
        crit_2 = dist_next > atol * ts_std

        fnn_prop[ind] = np.mean(np.logical_or(crit_1, crit_2))

    return fnn_prop


def embed_dim_cao(
    ts: np.ndarray,
    lag: int,
    dims: t.Union[int, t.Sequence[int]] = 16,
    ts_scaled: t.Optional[np.ndarray] = None,
) -> t.Tuple[np.ndarray, np.ndarray]:
    """Estimate Cao's metrics to estimate time-series embedding dimension.

    The Cao's metrics are two statistics, `E1` and `E2`, used to estimate the
    appropriate embedding metric of a time-series. From the `E1` statistic it
    can be defined the appropriate embedding dimension as the index after the
    saturation of the metric from a set of ordered lags.

    The precise `saturation` concept may be a subjective concept, since this
    metric can show some curious `artifacts` related to specific lags for
    specific time-series, which will need deeper further investigation.

    The `E2` statistics is to detect `false positives` from the `E1` statistic
    since if is used to distinguish between random white noise and a process
    generated from a true, non completely random, underlying process. If the
    time-series is purely random white noise, then all values of `E2` will be
    close to 1. If there exists a dimension with the `E2` metric estimated
    `sufficiently far` from 1, then this series is considered not a white
    random noise.

    Parameters
    ----------
    ts : :obj:`np.ndarray`
        One-dimensional time-series values.

    lag : int
        Embedding lag. You may want to check the `embed_lag` function
        documentation for embedding lag estimation. Must be a stricly
        positive value.

    dims : int or sequence of int
        Dimensions to estimate the Cao's `E1` and `E2` statistic values.
        If integer, estimate all dimensions from 1 up to the given number.
        If a sequence of integers, estimate all Cao's statistics for all
        given dimensions, and return the corresponding values in the same
        order of the given dimensions.
        All dimensions with non-positive values will receive a `np.nan`
        value for both Cao's metric.

    ts_scaled : :obj:`np.ndarray`, optional
        Standardized time-series values. Used to take advantage of
        precomputations.

    Returns
    -------
    tuple of :obj:`np.ndarray`
        `E1` and `E2` Cao's metrics, necessarily in that order, for all
        given dimensions (and with direct index correspondence for the
        given dimensions).

    References
    ----------
    .. [1] Liangyue Cao, Practical method for determining the minimum
        embedding dimension of a scalar time series, Physica D: Nonlinear
        Phenomena, Volume 110, Issues 1–2, 1997, Pages 43-50,
        ISSN 0167-2789, https://doi.org/10.1016/S0167-2789(97)00118-8.
    """
    if lag <= 0:
        raise ValueError("'lag' must be positive (got {}).".format(lag))

    _dims: t.Sequence[int]

    if np.isscalar(dims):
        _dims = np.arange(1, int(dims) + 1)  # type: ignore

    else:
        _dims = np.asarray(dims, dtype=int)

    ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

    ed, ed_star = np.zeros((2, len(_dims)), dtype=float)

    for ind, dim in enumerate(_dims):
        try:
            emb_next = embed_ts(ts=ts_scaled, lag=lag, dim=dim + 1)
            emb_cur = emb_next[:, 1:]

        except ValueError:
            ed[ind] = np.nan
            ed_star[ind] = np.nan
            continue

        nn_inds, dist_cur = nn(embed=emb_cur)

        emb_next_abs_diff = np.abs(emb_next[:, 0] - emb_next[nn_inds, 0])
        # Note: 'chebyshev'/'manhattan'/'L1'/max norm distance of X and Y,
        # both in the embed of (d + 1) dimensions, can be defined in respect
        # to one dimension less:
        # L1(X_{d+1}, Y_{d+1}) = |X_{d+1}, Y_{d+1}|_{inf}
        #   = max(|x_1 - y_1|, ..., |x_{d+1} - y_{d+1}|)
        #   = max(max(|x_1 - y_1|, ..., |x_d - y_d|), |x_{d+1} - y_{d+1}|)
        #   = max(L1(X_{d}, Y_{d}), |x_{d+1} - y_{d+1}|)
        dist_next = np.maximum(dist_cur, emb_next_abs_diff)

        # Note: 'ed' and 'ed_star' refers to, respectively, E_{d} and
        # E^{*}_{d} from the Cao's paper.
        ed[ind] = np.mean(dist_next / dist_cur)
        ed_star[ind] = np.mean(emb_next_abs_diff)

    # Note: the minimum embedding dimension is D such that e1[D]
    # is the first index where e1 stops changing significantly.
    e1 = ed[1:] / ed[:-1]

    # Note: This is the E2(d) Cao's metric. Its purpose is to
    # separate random time-series. For random-generated time-
    # series, e2 will be 1 for any dimension. For deterministic
    # data, however, e2 != 1 for some d.
    e2 = ed_star[1:] / ed_star[:-1]

    return e1, e2


def embed_lag(
    ts: np.ndarray,
    lag: t.Optional[t.Union[str, int]] = None,
    default_lag: int = 1,
    max_nlags: t.Optional[int] = None,
    detrended_acfs: t.Optional[np.ndarray] = None,
    detrended_ami: t.Optional[np.ndarray] = None,
    **kwargs
) -> int:
    """Find the appropriate embedding lag using a given criteria.

    Parameters
    ----------
    ts : :obj:`np.ndarray`
        One-dimensional time-series values.

    lag : int or str, optional (default = None)
        If scalar, return its own value casted to integer,

        If string, it must be one value in {`ami`, `acf`, `acf-nonsig`},
        which defines the strategy of defining the appropriate lag of
        the embedding.
            1. `ami`: uses the first minimum lag of the automutual information
                of the time-series.
            2. `acf`: uses the first negative lag of the autocorrelation of the
                time-series.
            3. `acf-nonsig` (default): uses the first non-significant lag of
                the time-series autocorrelation function. The non-significant
                value is defined as the first lag that has the absolute value
                of is autocorrelation below the critical value defined as
                1.96 / sqrt(ts.size).

        If None, the lag will be searched will the 'acf-nonsig'
        criteria.

    max_nlags : int, optional
        If ``lag`` is not a numeric value, than it will be estimated using
        either the time-series autocorrelation or mutual information
        function estimated up to this argument value.

    detrended_acfs : :obj:`np.ndarray`, optional
        Array of time-series autocorrelation function (for distinct ordered
        lags) of the detrended time-series. Used only if ``lag`` is any of
        `acf`, `acf-nonsig` or None.  If this argument is not given and the
        previous condiditon is meet, the autocorrelation function will be
        calculated inside this method up to ``max_nlags``.

    detrended_ami : :obj:`np.ndarray`, optional
        Array of time-series automutual information function (for distinct
        ordered lags). Used only if ``lag`` is `ami`. If not given and the
        previous condiditon is meet, the automutual information function
        will be calculated inside this method up to ``max_nlags``.

    kwargs:
        Extra arguments for the function used to estimate the lag. used
        only if `lag` is not a numeric value.

    Returns
    -------
    int
        Estimated embedding lag.

    Notes
    -----
    This method may be used to estimate `auto-interations` of the time-series
    (such as calculating the autocorrelation function, for instance) aswell.
    """
    VALID_OPTIONS = {
        "ami": info_theory.MFETSInfoTheory.ft_ami_first_critpt,
        "acf": autocorr.MFETSAutocorr.ft_acf_first_nonpos,
        "acf-nonsig": autocorr.MFETSAutocorr.ft_acf_first_nonsig,
    }  # type: t.Dict[str, t.Callable[..., t.Union[float, int]]]

    if lag is None:
        lag = "acf-nonsig"

    if isinstance(lag, str):
        if lag not in VALID_OPTIONS:
            raise ValueError(
                "'lag' must be in {} (got '{}')."
                "".format(VALID_OPTIONS.keys(), lag)
            )

        if max_nlags is None:
            max_nlags = ts.size // 2

        if lag == "ami":
            kwargs["detrended_ami"] = detrended_ami

        else:
            kwargs["detrended_acfs"] = detrended_acfs

        kwargs["max_nlags"] = max_nlags

        estimated_lag = VALID_OPTIONS[lag](ts, **kwargs)

        return default_lag if np.isnan(estimated_lag) else int(estimated_lag)

    if np.isscalar(lag):
        lag = int(lag)

        if lag <= 0:
            raise ValueError("'lag' must be positive (got {}).".format(lag))

        return lag

    raise TypeError(
        "'lag' type must be a scalar, a string or None (got {})."
        "".format(type(lag))
    )



"""Module dedicated to general time-series meta-features."""
import typing as t

import numpy as np
import scipy.spatial
import scipy.odr
import scipy.signal

import pymfe._embed as _embed
import pymfe._period as _period
import pymfe._utils as _utils


class MFETSGeneral:
    """Extract time-series meta-features from General group."""

    @classmethod
    def precompute_ts_scaled(
        cls, ts: np.ndarray, **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute a standardized time series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).
        """
        precomp_vals = {}  # type: t.Dict[str, np.ndarray]

        if "ts_scaled" not in kwargs:
            precomp_vals["ts_scaled"] = _utils.standardize_ts(ts=ts)

        return precomp_vals

    @classmethod
    def precompute_period(cls, ts: np.ndarray, **kwargs) -> t.Dict[str, int]:
        """Precompute the time-series period.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``ts_period`` (:obj:`int`): time-series period.
        """
        precomp_vals = {}  # type: t.Dict[str, int]

        if "ts_period" not in kwargs:
            precomp_vals["ts_period"] = cls.ft_period(ts=ts)

        return precomp_vals

    @classmethod
    def precompute_embed_caos_method(
        cls,
        ts: np.ndarray,
        dims: t.Union[int, t.Sequence[int]] = 16,
        lag: t.Optional[t.Union[str, int]] = None,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        emb_dim_cao_e1: t.Optional[np.ndarray] = None,
        emb_dim_cao_e2: t.Optional[np.ndarray] = None,
        **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute estimated Cao's method E1 and E2 values.

        Using the Cao's embedding dimension estimation, it is calculated both
        of its metrics, `E1` and `E2` whose purpose is to, respectively,
        detect the appropriate embedding dimension and detect whether the given
        time-series is generated by a completely random process (white noise).

        The appropriate embedding dimension is the saturation dimension of `E1`
        if and only if exists a dimension `E2` sufficiently distinct from 1.
        If `E2` is approximately constant at 1 over all dimensions, the series
        is considered white noise and, therefore, the embedding dimension is
        assumed to be 1.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        dims : int or a sequence of int, optional (default=16)
            The embedding dimension candidates. In int, investigate all values
            between 1 and ``dims`` value (both inclusive). If a sequence of
            integers is used, then investigate only the given dimensions.

        lag : int or str, optional
            Lag of the time-series embedding. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        emb_dim_cao_e1 : :obj:`np.ndarray`, optional
            E1 values from the Cao's method. Used to take advantage of
            precomputations.

        emb_dim_cao_e2 : :obj:`np.ndarray`, optional
            E2 values from the Cao's method. Used to take advantage of
            precomputations.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed items are returned:
                * ``emb_dim_cao_e1`` (:obj:`np.ndarray`): E1 values of the
                    Cao's method for all given dimensions.
                * ``emb_dim_cao_e2`` (:obj:`np.ndarray`): E2 values of the
                    Cao's method for all given dimensions.

            The following items are necessary and, therefore, also precomputed
            if necessary:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).
                * ``lag`` (int): estimative of the adequate embedding lag of
                    the time-series values.

        References
        ----------
        .. [1] Liangyue Cao, Practical method for determining the minimum
            embedding dimension of a scalar time series, Physica D: Nonlinear
            Phenomena, Volume 110, Issues 1–2, 1997, Pages 43-50,
            ISSN 0167-2789, https://doi.org/10.1016/S0167-2789(97)00118-8.
        """
        precomp_vals = {}  # type: t.Dict[str, np.ndarray]

        ts_scaled = kwargs.get("ts_scaled")

        if ts_scaled is None:
            precomp_vals.update(cls.precompute_ts_scaled(ts=ts))
            ts_scaled = precomp_vals["ts_scaled"]

        if lag is None or isinstance(lag, str):
            lag = _embed.embed_lag(
                ts=ts_scaled,
                detrended_acfs=detrended_acfs,
                detrended_ami=detrended_ami,
                max_nlags=max_nlags,
            )

            precomp_vals["lag"] = lag

        if emb_dim_cao_e1 is None or emb_dim_cao_e2 is None:
            emb_dim_cao_e1, emb_dim_cao_e2 = _embed.embed_dim_cao(
                ts=ts, ts_scaled=ts_scaled, dims=dims, lag=lag
            )

            precomp_vals["emb_dim_cao_e1"] = emb_dim_cao_e1
            precomp_vals["emb_dim_cao_e2"] = emb_dim_cao_e2

        return precomp_vals

    @classmethod
    def precompute_walker(
        cls,
        ts: np.ndarray,
        step_size: float = 0.1,
        start_point: t.Optional[t.Union[int, float]] = None,
        **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute the path of a particle attracted by the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        step_size : float, optional (default=0.1)
            Strength of the attractive force. In each time step, the particle
            position will be a weighted average of its previous position, with
            weight (1 - ``step_size``), and the current time-series value, with
            weight ``step_size``.

        start_point : int or float, optional
            Particle starting position. If None, will start at `0` (which is
            the mean value of the standardized time-series values).

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``walker_path`` (:obj:`np.ndarray`): path of a particle
                    attracted by the current time-series value, for a given
                    attractive strength (``step_size``) and initial position
                    (``start_point``).

            The following item is necessary and, therefore, also precomputed
            if necessary:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        precomp_vals = {}  # type: t.Dict[str, np.ndarray]

        if "walker_path" not in kwargs:
            ts_scaled = kwargs.get("ts_scaled")

            if ts_scaled is None:
                precomp_vals.update(cls.precompute_ts_scaled(ts=ts))
                ts_scaled = precomp_vals["ts_scaled"]

            walker_path = cls._ts_walker(
                ts=ts_scaled, step_size=step_size, start_point=start_point
            )

            precomp_vals["walker_path"] = walker_path

        return precomp_vals

    @staticmethod
    def _calc_season_mode_ind(
        ts_season: np.ndarray,
        ts_period: int,
        indfunc: t.Callable[[np.ndarray], float],
    ) -> int:
        """Calculate a mode index based on the time-series seasonality.

        Used by both ``ft_trough_frac`` and ``ft_peak_frac`` to calculate,
        respectively, the mode of the argmin and argmax for all seasons.
        """
        inds = np.arange(ts_period)

        inds = np.array(
            [
                indfunc(ts_season[i * ts_period + inds])
                for i in np.arange(ts_season.size // ts_period)
            ],
            dtype=int,
        )

        mode_inds, _ = scipy.stats.mode(inds)
        return mode_inds[0] + 1

    @classmethod
    def _ts_walker(
        cls,
        ts: np.ndarray,
        step_size: float = 0.1,
        start_point: t.Optional[t.Union[int, float]] = None,
    ) -> np.ndarray:
        """Simulate a particle attracted to the current time-series value.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if start_point is None:
            # Note: actually, it is used mean(ts) as starting point. However,
            # we are using ts_scaled and, therefore, mean(ts_scaled) = 0.
            start_point = 0.0

        walker_path = np.zeros(ts.size, dtype=float)
        walker_path[0] = start_point

        for i in np.arange(1, ts.size):
            diff = ts[i - 1] - walker_path[i - 1]
            # Note: weighted average between the current time-series value
            # (with 'step_size' weight) and the previous particle position
            # (with '1 - step_size' weight).
            walker_path[i] = walker_path[i - 1] + step_size * diff

        return walker_path

    @classmethod
    def ft_length(cls, ts: np.ndarray) -> int:
        """Length of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        Returns
        -------
        int
            Length of the time-seties.

        References
        ----------
        .. [1] Prudêncio, Ricardo & Ludermir, Teresa. (2004). Meta-learning
            approaches to selecting time series models. Neurocomputing. 61.
            121-137.  10.1016/j.neucom.2004.03.008.
        """
        return ts.size

    @classmethod
    def ft_diff(cls, ts: np.ndarray, order: int = 1) -> np.ndarray:
        """`n`th-order difference of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        order : int, optional (default=1)
            Order of differentiation.

        Returns
        -------
        :obj:`np.ndarray`
            `n`th-order differenced time-series values.
        """
        return np.diff(ts, n=order)

    @classmethod
    def ft_period(
        cls, ts: np.ndarray, ts_period: t.Optional[int] = None
    ) -> int:
        """Period of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        ts_period : int, optional
            Time-series period. Used to take advantage of precomputations.

        Returns
        -------
        int
            Time-series period.
        """
        return _period.get_ts_period(ts=ts, ts_period=ts_period)

    @classmethod
    def ft_turning_points(cls, ts: np.ndarray) -> np.ndarray:
        """Turning points in the time-series.

        A turning point is a time-series point `p_{i}` which both neighbor
        values, p_{i-1} and p_{i+1}, are either lower (p_{i} > p_{i+1} and
        p_{i} > p_{i-1}) or higher (p_{i} < p_{i+1} and p_{i} < p_{i-1}).

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        Returns
        -------
        :obj:`np.ndarray`
            Binary array marking (with `1`) where is the turning points in the
            time-series.

        References
        ----------
        .. [1] C. Shah, Model selection in univariate time series forecasting
            using discriminant analysis, International Journal of Forecasting
            13 (4) (1997) 489–500.
        .. [2] Prudêncio, Ricardo & Ludermir, Teresa. (2004). Meta-learning
            approaches to selecting time series models. Neurocomputing. 61.
            121-137.  10.1016/j.neucom.2004.03.008.
        .. [3] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        return _utils.find_crit_pt(ts, type_="non-plateau").astype(int)

    @classmethod
    def ft_turning_points_trend(cls, ts_trend: np.ndarray) -> np.ndarray:
        """Turning points in the time-series trend.

        A turning point is a time-series point `p_{i}` which both neighbor
        values, p_{i-1} and p_{i+1}, are either lower (p_{i} > p_{i+1} and
        p_{i} > p_{i-1}) or higher (p_{i} < p_{i+1} and p_{i} < p_{i-1}).

        Parameters
        ----------
        ts_trend : :obj:`np.ndarray`
            One-dimensional time-series trend values.

        Returns
        -------
        :obj:`np.ndarray`
            Binary array marking where is the turning points in the
            time-series trend.

        References
        ----------
        .. [1] C. Shah, Model selection in univariate time series forecasting
            using discriminant analysis, International Journal of Forecasting
            13 (4) (1997) 489–500.
        .. [2] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        return cls.ft_turning_points(ts=ts_trend)

    @classmethod
    def ft_step_changes(cls, ts: np.ndarray, ddof: int = 1) -> np.ndarray:
        """Step change points in the time-series.

        Let p_{t_{a}}^{t_{b}} be the subsequence of observations from the
        timestep t_{a} and t_{b}, both inclusive. A point `p_i` is a
        step change if and only if:

        abs(p_{i} - mean(p_{1}^{i-1})) > 2 * std(p_{1}^{i-1})

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        ddof : float, optional (default=1)
            Degrees of freedom for standard deviation.

        Returns
        -------
        :obj:`np.ndarray`
            Binary array marking where is the step change points in the
            time-series.

        References
        ----------
        .. [1] C. Shah, Model selection in univariate time series forecasting
            using discriminant analysis, International Journal of Forecasting
            13 (4) (1997) 489–500.
        .. [2] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        ts_cmeans = np.cumsum(ts) / np.arange(1, ts.size + 1)

        ts_mean_abs_div = np.abs(ts[1:] - ts_cmeans[:-1])

        step_changes = np.array(
            [
                int(ts_mean_abs_div[i - 1] > 2 * np.std(ts[:i], ddof=ddof))
                for i in np.arange(1 + ddof, ts.size)
            ],
            dtype=int,
        )

        return step_changes

    @classmethod
    def ft_step_changes_trend(
        cls, ts_trend: np.ndarray, ddof: int = 1
    ) -> np.ndarray:
        """Step change points in the time-series trend.

        Let p_{t_{a}}^{t_{b}} be the subsequence of observations from the
        timestep t_{a} and t_{b}, both inclusive. A point `p_i` is a
        step change if and only if:

        abs(p_{i} - mean(p_{1}^{i-1})) > 2 * std(p_{1}^{i-1})

        Parameters
        ----------
        ts_trend : :obj:`np.ndarray`
            One-dimensional time-series trend values.

        ddof : float, optional (default=1)
            Degrees of freedom for standard deviation.

        Returns
        -------
        :obj:`np.ndarray`
            Binary array marking where is the step change points in the
            time-series trend.

        References
        ----------
        .. [1] C. Shah, Model selection in univariate time series forecasting
            using discriminant analysis, International Journal of Forecasting
            13 (4) (1997) 489–500.
        .. [2] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        step_changes = cls.ft_step_changes(ts=ts_trend, ddof=ddof)
        return step_changes

    @classmethod
    def ft_pred(
        cls,
        ts: np.ndarray,
        embed_dim: int = 2,
        std_range: t.Union[int, float] = 3,
        num_spacing: t.Union[int, float] = 4,
        metric: str = "minkowski",
        p: t.Union[int, float] = 2,
        ddof: int = 1,
        lag: t.Optional[t.Union[int, str]] = None,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Signal predictability using delay vector variance method.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        embed_dim : int, optional (default=2)
            Dimension of the time-series embedding.

        std_range : int or float, optional (default=3)
            Number of standard deviations in range of the delay vector
            variance analysis (around the mean value).

        num_spacing : int, optional (default=4)
            Controls how fine is the spacing between the analysed values
            within the analysed range. The greater this parameter is, the
            less is the spacing between the analysed values around the mean.

        metric : str, optional (default="minkowski")
            Distance metric to measure the pairwise distance between embedded
            time-series instances. Check `scipy.spatial.distance.pdist`
            documentation to see the full list of available distance metrics.

        p : int or float, optional (default=2)
            Power parameter to minkowski metric (used only if metric is
            "minkowski").

        ddof : int, optional (default=1)
            Degrees of freedom to calculate the variance within this function.
            (standard deviation of pairwise distances of embedded time-seres
            and the variance of radius nearest neighbors from the definition
            of the implemented meta-feature. Check references for in-depth
            information.)

        lag : int or str, optional
            Lag of the time-series embedding. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        ts_scaled : :obj:`np.ndarray`
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            Array of predictability estimation of the embedded time-series
            using delay vector variance method.

        References
        ----------
        .. [1] Mandic DP, Chambers JA. 2001. Recurrent neural networks for
            prediction: learning algorithms, architectures and stability.
            New York, NY: John Wiley & Sons, Ltd.
        .. [2] Gautama T, Mandic DP, Hulle MMV. 2004. The delay vector variance
            method for detecting determinism and nonlinearity in time series.
            Physica D 190, 167–176.
        .. [3] Gautama T, Hulle MMV, Mandic DP. 2004. On the characterisation
            of the deterministic/stochastic and linear/nonlinear nature of time
            series. Technical Report DPM-04-05. Imperial College London.
        .. [4] Mandic DP, Chen M, Gautama T, Van Hulle MM, Constantinides A.
            2008. On the characterization of the deterministic/stochastic and
            linear/nonlinear nature of time series. Proc. R. Soc. A 464,
            1141–1160.
        .. [5] Jaksic V, Mandic DP, Ryan K, Basu B, Pakrashi V. A comprehensive
            study of the delay vector variance method for quantification of
            nonlinearity in dynamical systems. R Soc Open Sci.
            2016;3(1):150493. Published 2016 Jan 6. doi:10.1098/rsos.150493
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)
        lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            max_nlags=max_nlags,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
        )

        ts_embed = _embed.embed_ts(ts_scaled, lag=lag, dim=embed_dim)

        dist_mat = scipy.spatial.distance.pdist(ts_embed, metric=metric, p=p)

        dist_mean = np.mean(dist_mat)
        dist_std = np.std(dist_mat, ddof=ddof)

        dist_mat = scipy.spatial.distance.squareform(dist_mat)

        # Note: prevents the instance itself be considered its own neighbor.
        dist_mat[np.diag_indices_from(dist_mat)] = np.inf

        var_sets = np.zeros(num_spacing, dtype=float)

        for i in np.arange(num_spacing):
            threshold = max(
                0.0,
                dist_mean
                + std_range * dist_std * (i * 2 / (num_spacing - 1) - 1),
            )

            neighbors = dist_mat <= threshold

            for neigh_inds in neighbors:
                if np.sum(neigh_inds) > ddof:
                    var_sets[i] += np.var(ts_embed[neigh_inds, :], ddof=ddof)

        # Note: originally, this value is also normalize dy the time-series
        # variance but, as we are using the standardized time-series, its
        # variance is equal 1.
        var_sets /= num_spacing

        return var_sets

    @classmethod
    def ft_frac_cp(
        cls,
        ts: np.ndarray,
        threshold: t.Optional[t.Union[int, float]] = None,
        normalize: bool = True,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """Cross-points of a given horizontal line.

        This method calculates the fraction that the transition between
        two consecutive observations in the time-series crosses a horizontal
        line given by `y = threshold` over all possible crosses (i.e., if
        the time-series values alternate between higher and lower values
        relative to the given threhsold in all transitions).

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        threshold : int or float, optional
            Threshold to define the horizontal line in the form y = threshold.
            If None, the threshold will default to the time-series median.

        normalize : bool, optional (defualt = True)
            If True, return the fraction of actual crosses over all possible
            crosses. If False, return the number of crosses.

        ts_scaled : :obj:`np.ndarray`
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        int or float
            If `normalize` is True, return the fraction of transition crosses
            over all possible crosses given a fixed time-series length and a
            threshold. If `normalize` is False, return the number of crosses.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        if threshold is None:
            threshold = np.median(ts_scaled)

        higher_med = ts_scaled <= threshold
        num_cp = np.sum(np.logical_xor(higher_med[1:], higher_med[:-1]))

        if normalize:
            num_cp /= ts_scaled.size - 1

        return num_cp

    @classmethod
    def ft_bin_mean(cls, ts: np.ndarray) -> np.ndarray:
        """Check whether values are below or above the average value.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        Returns
        -------
        :obj:`np.ndarray`
            Binary array with the time-series length, marking with `1`
            observations above or equal the mean value, and `0` otherwise.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        return (ts >= np.mean(ts)).astype(int)

    @classmethod
    def ft_fs_len(
        cls, ts: np.ndarray, num_bins: int = 10, strategy: str = "equal-width"
    ) -> np.ndarray:
        """Lenght of discretized time-series values plateaus.

        The time-series is discretized using a histogram with number of bins
        equal to `num_bins` bins and its cuts are calculated using a method
        defined by the `strategy` parameter.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_bins : int, optional (default=10)
            Number of bins to discretize the time-series values.

        strategy : str, optional (default=`equal-width`)
            Strategy used to define the histogram bins. Must be either
            `equal-width` (bins with equal with) or `equiprobable` (bins
            with the same amount of observations within).

        Returns
        -------
        :obj:`np.ndarray`
            Length of each plateau in the discretized time-series values.

        References
        ----------
        .. [1] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [2] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        ts_disc = _utils.discretize(
            ts=ts, num_bins=num_bins, strategy=strategy
        )

        i = 1
        counter = 1
        fs_len = []  # type: t.List[int]

        while i < ts.size:
            if not np.isclose(ts_disc[i], ts_disc[i - 1]):
                fs_len.append(counter)
                counter = 1

            else:
                counter += 1

            i += 1

        return np.asarray(fs_len, dtype=float)

    @classmethod
    def ft_peak_frac(
        cls,
        ts_season: np.ndarray,
        ts_period: t.Optional[int] = None,
        normalize: bool = True,
    ) -> t.Union[int, float]:
        """Fraction of where the seasonal peak is localed within a period.

        This method returns the fraction (relative to the time-series period)
        of the mode of the peak localization within every complete period
        in the given time-series. For instance, if the peak tends to appear in
        the middle of a time-series complete period, this method will return
        `0.5`. If, in contrast, the peak tends to appear as the first period
        observation, this method returns `0.0`.

        If the time-series is not seasonal (i.e., the seasonal component is
        empty and its period is 1), then this method returns `np.nan`.

        Parameters
        ----------
        ts_season : :obj:`np.ndarray`
            One-dimensional time-series seasonal component values.

        ts_period : int, optional
            Period of the time series. If None, the period will be estimated
            as the lag corresponding to the absolute maximum of the time-series
            autocorrelation function.

        normalize : bool, optional (default=True)
            If False, the result will be the mode index of the seasonal peak.

        Returns
        -------
        int or float
            If ``ts_period`` is less or equal than 1, return :obj:`np.nan`.
            If `normalize` is True, return the fraction of the seasonal peak
            mode index over the time-series period. If `normalize` is False,
            return the seasonal peak mode index.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        _ts_period = _period.get_ts_period(ts=ts_season, ts_period=ts_period)

        if _ts_period <= 1:
            return np.nan

        ind_peak: t.Union[int, float]

        ind_peak = cls._calc_season_mode_ind(
            ts_season=ts_season, ts_period=_ts_period, indfunc=np.argmax
        )

        if normalize:
            ind_peak /= _ts_period

        return ind_peak

    @classmethod
    def ft_trough_frac(
        cls,
        ts_season: np.ndarray,
        ts_period: t.Optional[int] = None,
        normalize: bool = True,
    ) -> t.Union[int, float]:
        """Fraction of where the seasonal trough is localed within a period.

        This method returns the fraction (relative to the time-series period)
        of the mode of the trough localization within every complete period
        in the given time-series. For instance, if the trough tends to appear
        in the middle of a time-series complete period, this method will return
        `0.5`. If, in contrast, the trough tends to appear as the first period
        observation, this method returns `0.0`.

        If the time-series is not seasonal (i.e., the seasonal component is
        empty and its period is 1), then this method returns `np.nan`.

        Parameters
        ----------
        ts_season : :obj:`np.ndarray`
            One-dimensional time-series seasonal component values.

        ts_period : int, optional
            Period of the time series. If None, the period will be estimated
            as the lag corresponding to the absolute maximum of the time-series
            autocorrelation function.

        normalize : bool, optional (default=True)
            If False, the result will be the mode index of the seasonal trough.

        Returns
        -------
        int or float
            If ``ts_period`` is less or equal than 1, return :obj:`np.nan`.
            If `normalize` is True, return the fraction of the seasonal trough
            mode index over the time-series period. If `normalize` is False,
            return the seasonal trough mode index.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        _ts_period = _period.get_ts_period(ts=ts_season, ts_period=ts_period)

        if _ts_period <= 1:
            return np.nan

        ind_trough: t.Union[int, float]

        ind_trough = cls._calc_season_mode_ind(
            ts_season=ts_season, ts_period=_ts_period, indfunc=np.argmin
        )

        if normalize:
            ind_trough /= _ts_period

        return ind_trough

    @classmethod
    def ft_walker_path(
        cls,
        ts: np.ndarray,
        step_size: float = 0.1,
        start_point: t.Optional[t.Union[int, float]] = None,
        relative_dist: bool = True,
        walker_path: t.Optional[np.ndarray] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Path of a particle attracted by the time-series values.

        Simulate a particle, starting from ``start_point``, which follows the
        current time-series value with ``step_size`` attractive strength.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        step_size : float, optional (default=0.1)
            Strength of the attractive force. In each time step, the particle
            position will be a weighted average of its previous position, with
            weight (1 - ``step_size``), and the current time-series value, with
            weight ``step_size``.

        start_point : int or float, optional
            Particle starting position. If None, will start at `0` (which is
            the mean value of the standardized time-series values).

        relative_dist : bool, optional (default=True)
            If True, return the distance of the particle to the current value
            of the time-series for each time-step. If False, then return the
            particle position in each time step.

        walker_path : :obj:`np.ndarray`, optional
            The position of a particle attracted by the current time-series
            value. Used to take advantage of precomputations.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            If `relative_dist` is True, the distance of the particle to the
            current value of the time-series for each time-step. Otherwise, if
            `relative_dist` is False, then return the particle position in each
            time step.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        if walker_path is None:
            walker_path = cls._ts_walker(
                ts=ts_scaled, step_size=step_size, start_point=start_point
            )

        if relative_dist:
            return np.abs(walker_path - ts_scaled)

        return walker_path

    @classmethod
    def ft_walker_cross_frac(
        cls,
        ts: np.ndarray,
        step_size: float = 0.1,
        start_point: t.Optional[t.Union[int, float]] = None,
        normalize: bool = True,
        walker_path: t.Optional[np.ndarray] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """Fraction of crosses by a particle attracted by the time-series.

        Simulate a particle, starting from ``start_point``, which follows the
        current time-series value with ``step_size`` attractive strength, and
        then return the fraction of crosses between the particle position
        transitions and the time-series values transitions.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        step_size : float, optional (default=0.1)
            Strength of the attractive force. In each time step, the particle
            position will be a weighted average of its previous position, with
            weight (1 - ``step_size``), and the current time-series value, with
            weight ``step_size``.

        start_point : int or float, optional
            Particle starting position. If None, will start at `0` (which is
            the mean value of the standardized time-series values).

        normalize : bool, optional (default=True)
            If True, normalize the number of crosses by the maximum possible
            number of crosses (len(ts) - 1). If false, return the number of
            crosses.

        walker_path : :obj:`np.ndarray`, optional
            The position of a particle attracted by the current time-series
            value. Used to take advantage of precomputations.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            If `normalize` is True, return the fraction of crosses between
            the time-series values transitions and the particle position
            transitions over the total possible crosses. If `normalize` is
            False, then return the crosses number.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        if walker_path is None:
            walker_path = cls._ts_walker(
                ts=ts_scaled, step_size=step_size, start_point=start_point
            )

        cross_num = np.sum(
            (walker_path[:-1] - ts_scaled[:-1])
            * (walker_path[1:] - ts_scaled[1:])
            < 0
        )

        if normalize:
            cross_num /= walker_path.size - 1

        return cross_num

    @classmethod
    def ft_moving_threshold(
        cls,
        ts: np.ndarray,
        rate_absorption: float = 0.1,
        rate_decay: float = 0.1,
        relative: bool = False,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Simulate a dynamic threshold based on the time-series values.

        The threshold is run over the absolute values of the standardize
        time-series T. Each time the current value of T surpasses the current
        threshold value V, it absorbs a fraction (``rate_absorption``) of the
        current T value. Otherwise, T decays at ``rate_decay``.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        rate_absorption : float, optional (default=0.1)
            Rate of absorption of the current time-series absolute value
            when it surpasses the current threshold value. Must be a value
            in (0, 1) range.

        rate_decay : float, optional (default=0.1)
            Rate of decay of the threshold when the current time-series
            absolute value is not higher than the current threshold value.
            Must be a value in (0, 1) range.

        relative : bool, optional (default=False)
            If True, this method will return the threshold values subtracted by
            the current scaled time-series absolute values. Therefore, the
            return value is the threshold value relative to the time-series
            (absolute) value.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            If `relative` is False, return the obtained threshold values for
            each time step. If `relative` is True, return the threshold values
            with the time-series absolute values subtracted.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if not 0 < rate_decay < 1:
            raise ValueError(
                "'rate_decay' must be in (0, 1) (got {})."
                "".format(rate_decay)
            )

        if not 0 < rate_absorption < 1:
            raise ValueError(
                "'rate_absorption' must be in (0, 1) (got"
                " {}).".format(rate_absorption)
            )

        ts_scaled = np.abs(_utils.standardize_ts(ts=ts, ts_scaled=ts_scaled))

        # Note: threshold[0] = std(ts_scaled) = 1.0.
        threshold = np.ones(1 + ts.size, dtype=float)

        _ra = 1 + rate_absorption
        _rd = 1 - rate_decay

        for ind in np.arange(ts_scaled.size):
            if ts_scaled[ind] > threshold[ind]:
                # Absorb from the time series (absolute) values
                threshold[1 + ind] = _ra * ts_scaled[ind]
            else:
                # Decay the threshold
                threshold[1 + ind] = _rd * threshold[ind]

        if relative:
            # Note: ignore the first initial threshold
            return threshold[1:] - ts_scaled

        return threshold

    @classmethod
    def ft_embed_in_shell(
        cls,
        ts: np.ndarray,
        radii: t.Tuple[float, float] = (0.0, 1.0),
        embed_dim: int = 2,
        lag: t.Optional[t.Union[str, int]] = None,
        normalize: bool = True,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """Fraction of embedded time-series observations inside a hypershell.

        The hypershell is defined as the region between (and including the
        surface of) a hypersphere of radius R and a second hypersphere of
        radius r (also including its surface region), R > r.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        radii : tuple of floats, optional (default=(0, 1))
            Tuple with the radii of both hyperspheres, in the form `(r, R)`
            with `r` being the radius of the inner (smaller) hypersphere and
            `R` being the radius of the outer (larger) hypersphere. `R` must
            be stricly positive while `r` must be non-negative. Also, `R`
            must be strictly larger than `r`.

        embed_dim : int, optional (default=2)
            The embedding dimension.

        lag : int or str, optional
            Lag of the time-series embedding. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        normalize : bool, optional (default=True)
            If True, return the fraction of the points inside the hypershell.
            If False, return the number of points inside the hypershell.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        int or float
            if `normalize` is True, return the fraction of the points inside
            the hypershell. If `normalize` is False, return the number of
            points inside the hypershell.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        radius_inner, radius_outer = radii

        if radius_inner < 0:
            raise ValueError(
                "Inner radius must be non-negative (got {})."
                "".format(radius_inner)
            )

        if radius_outer <= 0:
            raise ValueError(
                "Outer radius must be positive (got {})."
                "".format(radius_outer)
            )

        if radius_outer <= radius_inner:
            raise ValueError(
                "Outer radius ({}) must be higher than the inner "
                "radius ({}).".format(radius_outer, radius_inner)
            )

        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
            max_nlags=max_nlags,
        )

        # Note: embed is given by x(t) = [x(t), x(t-1), ..., x(t-m+1)]^T
        embed = _embed.embed_ts(ts_scaled, dim=embed_dim, lag=lag)

        # Note: here we supposed that every embed forms a zero-centered
        # hypersphere using the regular hypersphere equation:
        # sqrt(x_{i}^2 + x^{i-1}^2 + ... + x^{i-m+1}) = Radius
        embed_radius = np.linalg.norm(embed, ord=2, axis=1)

        # Note: we can check if every embed is in the same zero-centered
        # hypershell because all hyperspheres embeds are also zero-centered.
        in_shape_num = np.sum(
            np.logical_and(
                radius_inner <= embed_radius, embed_radius <= radius_outer
            )
        )

        if normalize:
            in_shape_num /= embed_radius.size

        return in_shape_num

    @classmethod
    def ft_force_potential(
        cls,
        ts: np.ndarray,
        potential: str = "sine",
        params: t.Optional[t.Tuple[float, float, float]] = None,
        start_point: t.Optional[t.Tuple[float, float]] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Simulate a force potential function using the time-series as input.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        potential : {"sine", "dblwell"}, optional (default="sine")
            Name of the potential function. Must be either `sine` or
            `dblwell`, defined as:
                1. sine: F(x) = sin(x / alpha) / alpha
                2. dblwell: F(x) = alpha ** 2 * x - x ** 3

        params : tuple of floats, optional
            Parameters for the force potential function. Must be a tuple with
            three values in the form (alpha, friction, time_step) which are,
            respectively, the coefficient in the potential force function,
            the decay rate of the particle velocity, and the size of the time
            step between each potential measurement.

        start_point : tuple of floats, optional
            Tuple in the form (initial_position, initial_velocity). If None,
            both the initial position and velocity will be 0, with the
            reasoning that position=0 is the mean value of the standardized
            time-series.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            Position of the potential at every time step.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        alpha: float

        # Note: 'DEF_PARAM' is in the following form:
        # potential_name: (default_parameters, force_function)
        DEF_PARAM = dict(
            sine=((1, 1, 0.1), lambda x: np.sin(x / alpha) / alpha),
            dblwell=((2, 0.1, 0.1), lambda x: alpha ** 2 * x - x ** 3),
        )

        if potential not in DEF_PARAM:
            raise ValueError(
                "'potential' must be in {} (got '{}')."
                "".format(DEF_PARAM.keys(), potential)
            )

        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        alpha, fric, dt = DEF_PARAM[potential][0] if params is None else params
        f_force = DEF_PARAM[potential][1]

        pos, vel = np.zeros((2, ts_scaled.size), dtype=float)

        # Note: it is actually used (mean(ts), 0.0) as default start
        # point, but mean(ts_scaled) = 0.
        pos[0], vel[0] = (0.0, 0.0) if start_point is None else start_point

        for t_prev in np.arange(ts_scaled.size - 1):
            aux = f_force(pos[t_prev]) + ts_scaled[t_prev] - fric * vel[t_prev]
            pos[t_prev + 1] = pos[t_prev] + dt * vel[t_prev] + dt ** 2 * aux
            vel[t_prev + 1] = vel[t_prev] + dt * aux

            if np.isinf(pos[t_prev + 1]):
                raise ValueError("Potential trajectory diverged.")

        return pos

    @classmethod
    def ft_stick_angles(
        cls, ts: np.ndarray, ts_scaled: t.Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Angle between observations with the same signal in the time domain.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            Angles between time-series observations of the same signal.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """

        def calc_angles(arr: np.ndarray, inds: np.ndarray) -> np.ndarray:
            """Calculate the angle between adjacent ``arr`` indices values."""
            # Note: normalizing the indices in [0, 1] range to avoid too much
            # interference from the artificial timestamps.
            norm_factor = arr.size - 1
            tangent = np.diff(arr[inds]) / np.diff(inds)
            return np.arctan(norm_factor * tangent)

        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        is_nonneg = ts_scaled >= 0
        ang_pos = calc_angles(arr=ts_scaled, inds=np.flatnonzero(is_nonneg))
        ang_neg = calc_angles(arr=ts_scaled, inds=np.flatnonzero(~is_nonneg))

        angles = np.hstack((ang_pos, ang_neg))

        return angles

    @classmethod
    def ft_emb_lag(
        cls,
        ts: np.ndarray,
        lag: t.Optional[t.Union[str, int]] = None,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> int:
        """Estimated appropriate embedding lag.

        The lag is estimated using either the automutual information function
        or the autocorrelation function on the detrended time-series (using
        Friedman's Super Smoother).

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        lag : int or str, optional
            Must be None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

            If a positive integer, this method will return the given value by
            default. This is used to take advantage of precomputations.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        int
            Estimation of the appropriate embedding lag using the selected
            strategy.

        References
        ----------
        .. [1] Friedman, J. H. 1984, A variable span scatterplot smoother
            Laboratory for Computational Statistics, Stanford University
            Technical Report No. 5. Available at:
            https://www.slac.stanford.edu/pubs/slacpubs/3250/slac-pub-3477.pdf
            Accessed on May 12 2020.
        .. [2] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        _lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
            max_nlags=max_nlags,
        )
        return _lag

    @classmethod
    def ft_emb_dim_cao(
        cls,
        ts: np.ndarray,
        dims: t.Union[int, t.Sequence[int]] = 16,
        lag: t.Optional[t.Union[str, int]] = None,
        tol_threshold: float = 0.05,
        check_e2: bool = True,
        max_nlags: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        emb_dim_cao_e1: t.Optional[np.ndarray] = None,
        emb_dim_cao_e2: t.Optional[np.ndarray] = None,
    ) -> int:
        """Embedding dimension estimation using Cao's method.

        Using the Cao's embedding dimension estimation, it is calculated both
        of its metrics, `E1` and `E2` whose purpose is to, respectively,
        detect the appropriate embedding dimension and detect whether the given
        time-series is generated by a completely random process (white noise).

        The appropriate embedding dimension is the saturation dimension of `E1`
        if and only if exists a dimension `E2` sufficiently distinct from 1.
        If `E2` is approximately constant at 1 over all dimensions, the series
        is considered white noise and, therefore, the embedding dimension is
        assumed to be 1.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        dims : int or a sequence of int, optional (default=16)
            The embedding dimension candidates. In int, investigate all values
            between 1 and ``dims`` value (both inclusive). If a sequence of
            integers is used, then investigate only the given dimensions.

        lag : int or str, optional
            Lag of the time-series embedding. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        tol_threshold : float, optional (default=0.05)
            Tolerance threshold to defined the maximum absolute diference
            between two E1 values in order to assume saturation. This same
            threshold is the minimum absolute deviation that E2 values must
            have in order to be considered different than 1.

        check_e2 : bool, optional (default=True)
            If True, check if there exist a Cao's E2 value different than 1,
            and return 1 if this condition is not satisfied. If False, ignore
            the E2 values.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        emb_dim_cao_e1 : :obj:`np.ndarray`, optional
            E1 values from the Cao's method. Used to take advantage of
            precomputations.

        emb_dim_cao_e2 : :obj:`np.ndarray`, optional
            E2 values from the Cao's method. Used to take advantage of
            precomputations.

        Returns
        -------
        int
            Estimation of the appropriate embedding dimension using Cao's
            method.

        References
        ----------
        .. [1] Liangyue Cao, Practical method for determining the minimum
            embedding dimension of a scalar time series, Physica D: Nonlinear
            Phenomena, Volume 110, Issues 1–2, 1997, Pages 43-50,
            ISSN 0167-2789, https://doi.org/10.1016/S0167-2789(97)00118-8.
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
            max_nlags=max_nlags,
        )

        if emb_dim_cao_e1 is None or (check_e2 and emb_dim_cao_e2 is None):
            emb_dim_cao_e1, emb_dim_cao_e2 = _embed.embed_dim_cao(
                ts=ts, ts_scaled=ts_scaled, dims=dims, lag=lag
            )

        if (
            check_e2
            and emb_dim_cao_e2 is not None
            and np.all(np.abs(emb_dim_cao_e2 - 1) < tol_threshold)
        ):
            return 1

        e1_abs_diff = np.abs(np.diff(emb_dim_cao_e1))

        first_max_ind = 0

        try:
            first_max_ind = np.flatnonzero(e1_abs_diff <= tol_threshold)[0]

        except IndexError:
            pass

        return first_max_ind + 1

    @classmethod
    def ft_cao_e1(
        cls,
        ts: np.ndarray,
        dims: t.Union[int, t.Sequence[int]] = 16,
        lag: t.Optional[t.Union[str, int]] = None,
        max_nlags: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        emb_dim_cao_e1: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Estimated Cao's method E1 values.

        Using the Cao's embedding dimension estimation, it is calculated both
        of its metrics, `E1` and `E2` whose purpose is to, respectively,
        detect the appropriate embedding dimension and detect whether the given
        time-series is generated by a completely random process (white noise).

        The appropriate embedding dimension is the saturation dimension of `E1`
        if and only if exists a dimension `E2` sufficiently distinct from 1.
        If `E2` is approximately constant at 1 over all dimensions, the series
        is considered white noise and, therefore, the embedding dimension is
        assumed to be 1.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        dims : int or a sequence of int, optional (default=16)
            The embedding dimension candidates. In int, investigate all values
            between 1 and ``dims`` value (both inclusive). If a sequence of
            integers is used, then investigate only the given dimensions.

        lag : int or str, optional
            Lag of the time-series embedding. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        emb_dim_cao_e1 : :obj:`np.ndarray`, optional
            E1 values from the Cao's method. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            E1 values from Cao's method for all given dimensions.

        References
        ----------
        .. [1] Liangyue Cao, Practical method for determining the minimum
            embedding dimension of a scalar time series, Physica D: Nonlinear
            Phenomena, Volume 110, Issues 1–2, 1997, Pages 43-50,
            ISSN 0167-2789, https://doi.org/10.1016/S0167-2789(97)00118-8.
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
            max_nlags=max_nlags,
        )

        if emb_dim_cao_e1 is None:
            emb_dim_cao_e1, _ = _embed.embed_dim_cao(
                ts=ts, ts_scaled=ts_scaled, dims=dims, lag=lag
            )

        return emb_dim_cao_e1

    @classmethod
    def ft_cao_e2(
        cls,
        ts: np.ndarray,
        dims: t.Union[int, t.Sequence[int]] = 16,
        lag: t.Optional[t.Union[str, int]] = None,
        max_nlags: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        emb_dim_cao_e2: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Estimated Cao's method E2 values.

        Using the Cao's embedding dimension estimation, it is calculated both
        of its metrics, `E1` and `E2` whose purpose is to, respectively,
        detect the appropriate embedding dimension and detect whether the given
        time-series is generated by a completely random process (white noise).

        The appropriate embedding dimension is the saturation dimension of `E1`
        if and only if exists a dimension `E2` sufficiently distinct from 1.
        If `E2` is approximately constant at 1 over all dimensions, the series
        is considered white noise and, therefore, the embedding dimension is
        assumed to be 1.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        dims : int or a sequence of int, optional (default=16)
            The embedding dimension candidates. In int, investigate all values
            between 1 and ``dims`` value (both inclusive). If a sequence of
            integers is used, then investigate only the given dimensions.

        lag : int or str, optional
            Lag of the time-series embedding. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        emb_dim_cao_e2 : :obj:`np.ndarray`, optional
            E2 values from the Cao's method. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            E2 values from Cao's method for all given dimensions.

        References
        ----------
        .. [1] Liangyue Cao, Practical method for determining the minimum
            embedding dimension of a scalar time series, Physica D: Nonlinear
            Phenomena, Volume 110, Issues 1–2, 1997, Pages 43-50,
            ISSN 0167-2789, https://doi.org/10.1016/S0167-2789(97)00118-8.
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
            max_nlags=max_nlags,
        )

        if emb_dim_cao_e2 is None:
            _, emb_dim_cao_e2 = _embed.embed_dim_cao(
                ts=ts, ts_scaled=ts_scaled, dims=dims, lag=lag
            )

        return emb_dim_cao_e2

    @classmethod
    def ft_fnn_prop(
        cls,
        ts: np.ndarray,
        dims: t.Union[int, t.Sequence[int]] = 16,
        lag: t.Optional[t.Union[str, int]] = None,
        max_nlags: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Proportion of False Nearest Neighbors in the embedded time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        dims : int or a sequence of int, optional (default=16)
            The embedding dimension candidates. In int, investigate all values
            between 1 and ``dims`` value (both inclusive). If a sequence of
            integers is used, then investigate only the given dimensions.

        lag : int or str, optional
            Lag of the time-series embedding. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        Returns
        -------
        :obj:`np.ndarray`
            Proportion of False Nearest Neighbors for each given dimension.

        References
        ----------
        .. [1] Determining embedding dimension for phase-space reconstruction
            using a geometrical construction, Kennel, Matthew B. and Brown,
            Reggie and Abarbanel, Henry D. I., Phys. Rev. A, volume 45, 1992,
            American Physical Society.
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
            max_nlags=max_nlags,
        )

        fnn_prop = _embed.embed_dim_fnn(
            ts=ts, ts_scaled=ts_scaled, dims=dims, lag=lag
        )

        return fnn_prop



"""Module dedicated to model-based time-series meta-features."""
import typing as t
import warnings

import sklearn.gaussian_process
import sklearn.preprocessing
import statsmodels.tsa.holtwinters
import statsmodels.tsa.arima.model
import statsmodels.regression
import statsmodels.tools
import numpy as np

import pymfe._utils as _utils
import pymfe._orthopoly as _orthopoly
import pymfe._period as _period


class MFETSModelBased:
    """Extract time-series meta-features from Model-Based group."""

    @classmethod
    def precompute_ts_scaled(
        cls, ts: np.ndarray, **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute a standardized time series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).
        """
        precomp_vals = {}  # type: t.Dict[str, np.ndarray]

        if "ts_scaled" not in kwargs:
            precomp_vals["ts_scaled"] = _utils.standardize_ts(ts=ts)

        return precomp_vals

    @classmethod
    def precompute_period(cls, ts: np.ndarray, **kwargs) -> t.Dict[str, int]:
        """Precompute the time-series period.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``ts_period`` (:obj:`int`): time-series period.
        """
        precomp_vals = {}  # type: t.Dict[str, int]

        if "ts_period" not in kwargs:
            precomp_vals["ts_period"] = _period.get_ts_period(ts=ts)

        return precomp_vals

    @classmethod
    def precompute_model_ets(
        cls,
        ts: np.ndarray,
        damped_trend: bool = False,
        ts_period: t.Optional[int] = None,
        **kwargs
    ) -> t.Dict[str, t.Any]:
        """Precompute a standardized time series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            If True, the exponential smoothing models will be have damping
            effects.

        ts_period : int, optional
            Time-series period. Used to take advantage of precomputations.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``res_model_des`` (:obj:`HoltWintersResultsWrapper`):
                    Double exponential smoothing model (exponential smoothing
                    model without the seasonal component) with additive trend
                    results.
                * ``res_model_ets`` (:obj:`HoltWintersResultsWrapper`):
                    Triple exponential smoothing model (exponential smoothing
                    model with the seasonal component) with additive components
                    results.

            The following items are necessary and, therefore, also precomputed
            if necessary:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).
                * ``ts_period`` (:obj:`int`): time-series period.

        References
        ----------
        .. [1] Winters, Peter R. Forecasting Sales by Exponentially Weighted
            Moving Averages, 1960, INFORMS, Linthicum, MD, USA
            https://doi.org/10.1287/mnsc.6.3.324
        .. [2] Charles C. Holt, Forecasting seasonals and trends by
            exponentially weighted moving averages, International Journal of
            Forecasting, Volume 20, Issue 1, 2004, Pages 5-10, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2003.09.015.
        """
        precomp_vals = {}  # type: t.Dict[str, t.Any]

        ts_scaled = kwargs.get("ts_scaled")

        if ts_scaled is None:
            precomp_vals.update(cls.precompute_ts_scaled(ts=ts))
            ts_scaled = precomp_vals["ts_scaled"]

        if "res_model_des" not in kwargs:
            model = cls._fit_res_model_des(
                ts=ts_scaled, damped_trend=damped_trend
            )
            precomp_vals["res_model_des"] = model

        if ts_period is None:
            precomp_vals.update(cls.precompute_period(ts=ts))
            ts_period = precomp_vals["ts_period"]

        if "res_model_ets" not in kwargs:
            model = cls._fit_res_model_ets(
                ts=ts_scaled, ts_period=ts_period, damped_trend=damped_trend
            )
            precomp_vals["res_model_ets"] = model

        return precomp_vals

    @classmethod
    def precompute_ioe_std_linear_model(
        cls, ts: np.ndarray, step_size: float = 0.05, **kwargs
    ) -> t.Dict[str, t.Any]:
        """Precompute linear model of std with iterative outlier exclusion.

        In the iterative outlier exclusion, a uniformly spaced set of
        thresholds over the time-series range is build in increasing order.
        For each threshold it is calculated a statistic of the timestamp
        values of instances larger or equal than the current threshold.

        This precomputed linear model is the standard deviation of the
        timestamps regressed on the thresholds of each iteration.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        step_size : float, optional (default=0.05)
            Increase of the outlier threshold in each iteration. Must be a
            number strictly positive.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``res_ioe_std_linreg`` (:obj:`RegressionResults`): Results
                    from the linear regression model of iterative outlier
                    exclusion timestamps standard deviation regression on the
                    thresholds.

            The following item is necessary and, therefore, also precomputed
            if necessary:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        precomp_vals = {}  # type: t.Dict[str, t.Any]

        ts_scaled = kwargs.get("ts_scaled")

        if ts_scaled is None:
            precomp_vals.update(cls.precompute_ts_scaled(ts=ts))
            ts_scaled = precomp_vals["ts_scaled"]

        if "res_ioe_std_linreg" not in kwargs:
            lin_reg_res = cls._fit_ioe_std_lin_model(
                ts=ts, ts_scaled=ts_scaled, step_size=step_size
            )
            precomp_vals["res_ioe_std_linreg"] = lin_reg_res

        return precomp_vals

    @staticmethod
    def _fit_ioe_std_lin_model(
        ts: np.ndarray,
        step_size: float = 0.05,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> statsmodels.regression.linear_model.RegressionResults:
        """Fit a linear model of IOE standard deviations onto thresholds.

        In the iterative outlier exclusion, a uniformly spaced set of
        thresholds over the time-series range is build in increasing order.
        For each threshold it is calculated a statistic of the timestamp
        values of instances larger or equal than the current threshold.

        This precomputed linear model is the standard deviation of the
        timestamps regressed on the thresholds of each iteration.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        step_size : float, optional (default=0.05)
            Increase of the outlier threshold in each iteration. Must be a
            number strictly positive.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`statsmodels.regression.linear_model.RegressionResults`
            Results from the linear regression model of iterative outlier
            exclusion timestamps standard deviation regression on the
            thresholds.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """

        def ioe_std_func(arr: np.ndarray) -> float:
            """Normalized standard deviation with ddof=1."""
            return np.std(arr, ddof=1) / np.sqrt(arr.size)

        ioe_std = _utils.calc_ioe_stats(
            ts=ts,
            funcs=[ioe_std_func],
            ts_scaled=ts_scaled,
            step_size=step_size,
        )

        thresholds = statsmodels.tools.add_constant(
            np.arange(ioe_std.size) * step_size
        )

        lin_reg_res = statsmodels.regression.linear_model.OLS(
            ioe_std, thresholds
        ).fit()

        return lin_reg_res

    @staticmethod
    def _fit_res_model_des(
        ts: np.ndarray,
        damped_trend: bool = False,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper:
        """Fit a double exponential smoothing model with additive trend.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            Whether or not the exponential smoothing model should include a
            damping component.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper`
            Results of a optimized double exponential smoothing model.

        References
        ----------
        .. [1] Holt, C. E. (1957). Forecasting seasonals and trends by
            exponentially weighted averages (O.N.R. Memorandum No. 52).
            Carnegie Institute of Technology, Pittsburgh USA.
            https://doi.org/10.1016/j.ijforecast.2003.09.015
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                module="statsmodels",
                category=statsmodels.tools.sm_exceptions.ConvergenceWarning,
            )

            model = statsmodels.tsa.holtwinters.ExponentialSmoothing(
                endog=ts_scaled,
                initialization_method="estimated",
                trend="additive",
                damped_trend=damped_trend,
                seasonal=None,
            ).fit()

        return model

    @staticmethod
    def _fit_res_model_ets(
        ts: np.ndarray,
        damped_trend: bool = False,
        grid_search_guess: bool = True,
        ts_period: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
    ) -> statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper:
        """Fit a triple exponential smoothing model with additive components.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            Whether or not the exponential smoothing model should include a
            damping component.

        grid_search_guess : bool, optional (default=True)
            If True, used grid search (a.k.a. brute force) to search for good
            starting parameters. If False, this method becomes more less
            computationally intensive, but may fail to converge with higher
            chances.

        ts_period : int, optional
            Time-series period.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper`
            Results of a optimized triple exponential smoothing model.

        References
        ----------
        .. [1] Winters, Peter R. Forecasting Sales by Exponentially Weighted
            Moving Averages, 1960, INFORMS, Linthicum, MD, USA
            https://doi.org/10.1287/mnsc.6.3.324
        .. [2] Charles C. Holt, Forecasting seasonals and trends by
            exponentially weighted moving averages, International Journal of
            Forecasting, Volume 20, Issue 1, 2004, Pages 5-10, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2003.09.015.
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        ts_period = _period.get_ts_period(ts=ts_scaled, ts_period=ts_period)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                module="statsmodels",
                category=statsmodels.tools.sm_exceptions.ConvergenceWarning,
            )

            model = statsmodels.tsa.holtwinters.ExponentialSmoothing(
                endog=ts_scaled,
                initialization_method="estimated",
                trend="additive",
                seasonal="additive",
                damped_trend=damped_trend,
                seasonal_periods=ts_period,
            ).fit(use_brute=grid_search_guess)

        return model

    @staticmethod
    def _fit_ortho_pol_reg(
        ts_trend: np.ndarray, degree: int = 2
    ) -> statsmodels.regression.linear_model.RegressionResults:
        """Regress the time-series trend on orthogonal polinomials.

        Parameters
        ----------
        ts_trend : :obj:`np.ndarray`
            One-dimensional time-series trend component.

        degree : int, optional (default=2)
            Degree of the highest order polynomial (and, therefore, the number
            of distinct polynomials used).

        Returns
        -------
        :obj:`statsmodels.regression.linear_model.RegressionResults`
            Optimized parameters of the linear model of the time-series trend
            component regressed on the orthogonal polynomials.
        """
        X = _orthopoly.ortho_poly(
            ts=np.linspace(0, 1, ts_trend.size),
            degree=degree,
            return_coeffs=False,
        )

        X = statsmodels.tools.add_constant(X)

        ts_trend_scaled = _utils.standardize_ts(ts=ts_trend)

        return statsmodels.regression.linear_model.OLS(
            ts_trend_scaled, X
        ).fit()

    @classmethod
    def ft_des_level(
        cls,
        ts: np.ndarray,
        damped_trend: bool = False,
        ts_scaled: t.Optional[np.ndarray] = None,
        res_model_des: t.Optional[
            statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper
        ] = None,
    ) -> float:
        """Double exponential smoothing model (additive trend) level parameter.

        The `level` parameter is also known as the `alpha` parameter, from the
        traditional Double Exponential Smoothing definition formula. It is the
        `smoothing` factor of the model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            Whether or not the exponential smoothing model should include a
            damping component.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        res_model_des : :obj:`HoltWintersResultsWrapper`, optional
            Results after fitting a double exponential smoothing model. Used
            to take advantage of precomputations.

        Returns
        -------
        float
            `Level` (or `alpha`) parameter from a double exponential smoothing
            model.

        References
        ----------
        .. [1] Winters, Peter R. Forecasting Sales by Exponentially Weighted
            Moving Averages, 1960, INFORMS, Linthicum, MD, USA
            https://doi.org/10.1287/mnsc.6.3.324
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        if res_model_des is None:
            res_model_des = cls._fit_res_model_des(
                ts=ts, ts_scaled=ts_scaled, damped_trend=damped_trend
            )

        param_level = res_model_des.params["smoothing_level"]

        return param_level

    @classmethod
    def ft_des_trend(
        cls,
        ts: np.ndarray,
        damped_trend: bool = False,
        ts_scaled: t.Optional[np.ndarray] = None,
        res_model_des: t.Optional[
            statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper
        ] = None,
    ) -> float:
        """Double exponential smoothing model (additive trend) trend parameter.

        The `trend` parameter is also known as the `beta` parameter, from the
        traditional Double Exponential Smoothing definition formula. This
        parameter controls the decay of the influence of the trend change into
        the model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            Whether or not the exponential smoothing model should include a
            damping component.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        res_model_des : :obj:`HoltWintersResultsWrapper`, optional
            Results after fitting a double exponential smoothing model. Used
            to take advantage of precomputations.

        Returns
        -------
        float
            `Trend` (or `beta`) parameter from a double exponential smoothing
            model.

        References
        ----------
        .. [1] Winters, Peter R. Forecasting Sales by Exponentially Weighted
            Moving Averages, 1960, INFORMS, Linthicum, MD, USA
            https://doi.org/10.1287/mnsc.6.3.324
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        if res_model_des is None:
            res_model_des = cls._fit_res_model_des(
                ts=ts, ts_scaled=ts_scaled, damped_trend=damped_trend
            )

        param_trend = res_model_des.params["smoothing_trend"]

        return param_trend

    @classmethod
    def ft_ets_level(
        cls,
        ts: np.ndarray,
        damped_trend: bool = True,
        ts_period: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        res_model_ets: t.Optional[
            statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper
        ] = None,
    ) -> float:
        """ETS (additive components) model level parameter.

        ETS models are also known as `Holt-Winters Exponential Smoothing`.

        The `level` parameter is also known as the `alpha` parameter, from the
        traditional Triple Exponential Smoothing definition formula. It is the
        `smoothing` factor of the model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            Whether or not the exponential smoothing model should include a
            damping component.

        ts_period : int, optional
            Time-series period.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        res_model_ets : :obj:`HoltWintersResultsWrapper`, optional
            Results after fitting a triple exponential smoothing model. Used
            to take advantage of precomputations.

        Returns
        -------
        float
            `Level` (or `alpha`) parameter from a ETS model.

        References
        ----------
        .. [1] Winters, Peter R. Forecasting Sales by Exponentially Weighted
            Moving Averages, 1960, INFORMS, Linthicum, MD, USA
            https://doi.org/10.1287/mnsc.6.3.324
        .. [2] Charles C. Holt, Forecasting seasonals and trends by
            exponentially weighted moving averages, International Journal of
            Forecasting, Volume 20, Issue 1, 2004, Pages 5-10, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2003.09.015.
        .. [3] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [4] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        if res_model_ets is None:
            res_model_ets = cls._fit_res_model_ets(
                ts=ts,
                ts_scaled=ts_scaled,
                ts_period=ts_period,
                damped_trend=damped_trend,
            )

        param_level = res_model_ets.params["smoothing_level"]

        return param_level

    @classmethod
    def ft_ets_trend(
        cls,
        ts: np.ndarray,
        damped_trend: bool = True,
        ts_period: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        res_model_ets: t.Optional[
            statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper
        ] = None,
    ) -> float:
        """ETS (additive components) model trend parameter.

        ETS models are also known as `Holt-Winters Exponential Smoothing`.

        The `trend` parameter is also known as the `beta` parameter, from the
        traditional Triple Exponential Smoothing definition formula. This
        parameter controls the decay of the influence of the trend change into
        the model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            Whether or not the exponential smoothing model should include a
            damping component.

        ts_period : int, optional
            Time-series period.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        res_model_ets : :obj:`HoltWintersResultsWrapper`, optional
            Results after fitting a triple exponential smoothing model. Used
            to take advantage of precomputations.

        Returns
        -------
        float
            `Trend` (or `beta`) parameter from a ETS model.

        References
        ----------
        .. [1] Winters, Peter R. Forecasting Sales by Exponentially Weighted
            Moving Averages, 1960, INFORMS, Linthicum, MD, USA
            https://doi.org/10.1287/mnsc.6.3.324
        .. [2] Charles C. Holt, Forecasting seasonals and trends by
            exponentially weighted moving averages, International Journal of
            Forecasting, Volume 20, Issue 1, 2004, Pages 5-10, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2003.09.015.
        .. [3] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [4] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        if res_model_ets is None:
            res_model_ets = cls._fit_res_model_ets(
                ts=ts,
                ts_scaled=ts_scaled,
                ts_period=ts_period,
                damped_trend=damped_trend,
            )

        param_trend = res_model_ets.params["smoothing_trend"]

        return param_trend

    @classmethod
    def ft_ets_season(
        cls,
        ts: np.ndarray,
        damped_trend: bool = True,
        ts_period: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        res_model_ets: t.Optional[
            statsmodels.tsa.holtwinters.results.HoltWintersResultsWrapper
        ] = None,
    ) -> float:
        """ETS (additive components) model seasonal parameter.

        ETS models are also known as `Holt-Winters Exponential Smoothing`.

        The `seasonal` parameter is also known as the `gamma` parameter, from
        the traditional Triple Exponential Smoothing definition formula. It
        controls the influence of the seasonal component into the model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        damped_trend : bool, optional (default=False)
            Whether or not the exponential smoothing model should include a
            damping component.

        ts_period : int, optional
            Time-series period.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        res_model_ets : :obj:`HoltWintersResultsWrapper`, optional
            Results after fitting a triple exponential smoothing model. Used
            to take advantage of precomputations.

        Returns
        -------
        float
            `Seasonal` (or `gamma`) parameter from a ETS model.

        References
        ----------
        .. [1] Winters, Peter R. Forecasting Sales by Exponentially Weighted
            Moving Averages, 1960, INFORMS, Linthicum, MD, USA
            https://doi.org/10.1287/mnsc.6.3.324
        .. [2] Charles C. Holt, Forecasting seasonals and trends by
            exponentially weighted moving averages, International Journal of
            Forecasting, Volume 20, Issue 1, 2004, Pages 5-10, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2003.09.015.
        .. [3] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [4] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        if res_model_ets is None:
            res_model_ets = cls._fit_res_model_ets(
                ts=ts,
                ts_scaled=ts_scaled,
                ts_period=ts_period,
                damped_trend=damped_trend,
            )

        param_season = res_model_ets.params["smoothing_seasonal"]

        return param_season

    @classmethod
    def ft_linearity(
        cls,
        ts_trend: np.ndarray,
        res_model_orthoreg: t.Optional[
            statsmodels.regression.linear_model.RegressionResults
        ] = None,
    ) -> float:
        """Linearity measure from a orthogonal polynomial linear regression.

        The linearity measure is defined as the coefficient associated with
        the first order orthogonal polynomial from a linear regression of the
        time-series trend component onto a pair of orthogonal polynomials of
        order 1 and 2.

        Parameters
        ----------
        ts_trend : :obj:`np.ndarray`
            One-dimensional time-series trend component.

        res_model_orthoreg : :obj:`RegressionResults`, optional
            Linear regression results from the time-series trend component
            regressed onto a pair of orthogonal polynomials of order 1 and
            2. Used to take advantage of precomputations.

        Returns
        -------
        float
            Linearity measure.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        if res_model_orthoreg is None:
            res_model_orthoreg = cls._fit_ortho_pol_reg(ts_trend=ts_trend)

        _, linearity, _ = res_model_orthoreg.params

        return linearity

    @classmethod
    def ft_curvature(
        cls,
        ts_trend: np.ndarray,
        res_model_orthoreg: t.Optional[
            statsmodels.regression.linear_model.RegressionResults
        ] = None,
    ) -> float:
        """Curvature measure from a orthogonal polynomial linear regression.

        The curvature measure is defined as the coefficient associated with
        the second order orthogonal polynomial from a linear regression of the
        time-series trend component onto a pair of orthogonal polynomials of
        order 1 and 2.

        Parameters
        ----------
        ts_trend : :obj:`np.ndarray`
            One-dimensional time-series trend component.

        res_model_orthoreg : :obj:`RegressionResults`, optional
            Linear regression results from the time-series trend component
            regressed onto a pair of orthogonal polynomials of order 1 and
            2. Used to take advantage of precomputations.

        Returns
        -------
        float
            Curvature measure.

        References
        ----------
        .. [1] R. J. Hyndman, E. Wang and N. Laptev, "Large-Scale Unusual Time
            Series Detection," 2015 IEEE International Conference on Data
            Mining Workshop (ICDMW), Atlantic City, NJ, 2015, pp. 1616-1619,
            doi: 10.1109/ICDMW.2015.104.
        .. [2] Hyndman, R. J., Wang, E., Kang, Y., & Talagala, T. (2018).
            tsfeatures: Time series feature extraction. R package version 0.1.
        .. [3] Pablo Montero-Manso, George Athanasopoulos, Rob J. Hyndman,
            Thiyanga S. Talagala, FFORMA: Feature-based forecast model
            averaging, International Journal of Forecasting, Volume 36, Issue
            1, 2020, Pages 86-92, ISSN 0169-2070,
            https://doi.org/10.1016/j.ijforecast.2019.02.011.
        """
        if res_model_orthoreg is None:
            res_model_orthoreg = cls._fit_ortho_pol_reg(ts_trend=ts_trend)

        _, _, curvature = res_model_orthoreg.params

        return curvature

    @classmethod
    def ft_avg_cycle_period(cls, ts: np.ndarray) -> float:
        r"""Average cycle period from a AR(2) model.

        The average cycle period is defined as:
        $$
            \frac{2 * pi}{\arccos{-4 * phi_1 * (1 - \phi_2) / \phi_2}}
        $$
        If and only if $phi_1^{2} + 4 * phi_2 < 0$, where $\phi_1$ and
        $\phi_2$ are the parameters of the AR model associated with,
        respectively, the $y_{t-1}$ and $y_{t-2}$ previous observations
        from a time series $y$.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        Returns
        -------
        float
            Time-series average cycle period, if any. Return `np.nan` if no
            cycle is detected by the AR(2) model.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on May 23 2020.
            URL to this fomula: https://otexts.com/fpp2/non-seasonal-arima.html
        """
        model_res = statsmodels.tsa.arima.model.ARIMA(
            ts, order=(2, 0, 0)
        ).fit()
        theta_a, theta_b = model_res.arparams

        has_cycle = theta_a ** 2 + 4 * theta_b < 0

        if not has_cycle:
            return np.nan

        avg_cycle_period = (
            2 * np.pi / np.arccos(-0.25 * theta_a * (1 - theta_b) / theta_b)
        )

        return avg_cycle_period

    @classmethod
    def ft_gaussian_r_sqr(
        cls,
        ts: np.ndarray,
        random_state: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        gaussian_model: t.Optional[
            sklearn.gaussian_process.GaussianProcessRegressor
        ] = None,
    ) -> float:
        """R^2 from a gaussian process model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        random_state : int, optional
            Random seed to optimize the gaussian process model, to keep
            the results reproducible.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        gaussian_model : :obj:`GaussianProcessRegressor`, optional
            A fitted model of a gaussian process. Used to take advantage of
            precomputations.

        Returns
        -------
        float
            R^2 of a gaussian process model.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        gaussian_model = _utils.fit_gaussian_process(
            ts=ts_scaled,
            random_state=random_state,
            gaussian_model=gaussian_model,
            ts_scaled=ts_scaled,
        )

        X = np.linspace(0, 1, ts_scaled.size).reshape(-1, 1)
        r_squared = gaussian_model.score(X=X, y=ts_scaled)

        return r_squared

    @classmethod
    def ft_ioe_std_slope(
        cls,
        ts: np.ndarray,
        step_size: float = 0.05,
        ts_scaled: t.Optional[np.ndarray] = None,
        res_ioe_std_linreg: t.Optional[
            statsmodels.regression.linear_model.RegressionResults
        ] = None,
    ) -> float:
        """Linear model of IOE standard deviations onto thresholds slope.

        In the iterative outlier exclusion, a uniformly spaced set of
        thresholds over the time-series range is build in increasing order.
        For each threshold it is calculated a statistic of the timestamp
        values of instances larger or equal than the current threshold.

        This method calculates the slope coefficient of a linear model of the
        standard deviation of the timestamps regressed on the thresholds of
        each iteration.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        step_size : float, optional (default=0.05)
            Increase of the outlier threshold in each iteration. Must be a
            number strictly positive.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        res_ioe_std_linreg : :obj:`RegressionResults`, optional
            Results from the linear regression model of iterative outlier
            exclusion timestamps standard deviation regression on the
            thresholds.

        Returns
        -------
        float
            Slope from the linear regression model.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if res_ioe_std_linreg is None:
            res_ioe_std_linreg = cls._fit_ioe_std_lin_model(
                ts=ts, step_size=step_size, ts_scaled=ts_scaled
            )

        _, slope = res_ioe_std_linreg.params

        return slope

    @classmethod
    def ft_ioe_std_adj_r_sqr(
        cls,
        ts: np.ndarray,
        step_size: float = 0.05,
        ts_scaled: t.Optional[np.ndarray] = None,
        res_ioe_std_linreg: t.Optional[
            statsmodels.regression.linear_model.RegressionResults
        ] = None,
    ) -> float:
        """Linear model of IOE standard deviations onto thresholds Adj. R^2.

        In the iterative outlier exclusion, a uniformly spaced set of
        thresholds over the time-series range is build in increasing order.
        For each threshold it is calculated a statistic of the timestamp
        values of instances larger or equal than the current threshold.

        This method calculates the Adjusted R^2 from the linear model of the
        standard deviation of the timestamps regressed on the thresholds of
        each iteration.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        step_size : float, optional (default=0.05)
            Increase of the outlier threshold in each iteration. Must be a
            number strictly positive.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations.

        res_ioe_std_linreg : :obj:`RegressionResults`, optional
            Results from the linear regression model of iterative outlier
            exclusion timestamps standard deviation regression on the
            thresholds.

        Returns
        -------
        float
            Adjusted R^2 from the linear regression model.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if res_ioe_std_linreg is None:
            res_ioe_std_linreg = cls._fit_ioe_std_lin_model(
                ts=ts, step_size=step_size, ts_scaled=ts_scaled
            )

        adj_r_sqr = res_ioe_std_linreg.rsquared_adj

        return adj_r_sqr



"""Module dedicated to frequency domain time-series meta-features."""
import typing as t

import numpy as np
import scipy.signal
import scipy.stats


class MFETSFreqDomain:
    """Extract time-series meta-features from Frequency Domain group."""

    @classmethod
    def precompute_ps_residuals(
        cls, ts_residuals: np.ndarray, **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute the power spectrum frequency of the residuals.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals of an one-dimensional time-series.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``ps_residuals`` (:obj:`np.ndarray`): power spectrum
                    frequency of the time-series residuals.

        References
        ----------
        .. [1] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time averaging
            over short, modified periodograms", IEEE Trans. Audio
            Electroacoust. vol. 15, pp. 70-73, 1967.
        """
        precomp_vals = {}  # type: t.Dict[str, np.ndarray]

        if "ps_residuals" not in kwargs:
            ps_residuals = cls._calc_ps_residuals(ts_residuals=ts_residuals)
            precomp_vals["ps_residuals"] = ps_residuals

        return precomp_vals

    @classmethod
    def _calc_ps_residuals(
        cls,
        ts_residuals: np.ndarray,
        window: str = "hamming",
        scaling: str = "spectrum",
        return_freqs: bool = False,
    ) -> t.Union[np.ndarray, t.Tuple[np.ndarray, np.ndarray]]:
        """Calculate the positive side power spectrum of a fourier signal.

        References
        ----------
        .. [1] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time averaging
            over short, modified periodograms", IEEE Trans. Audio
            Electroacoust. vol. 15, pp. 70-73, 1967.
        """
        freqs, ps = scipy.signal.periodogram(
            ts_residuals,
            detrend=None,
            window=window,
            scaling=scaling,
            return_onesided=True,
        )

        if return_freqs:
            return freqs, ps

        return ps

    @classmethod
    def ft_ps_residuals(
        cls,
        ts_residuals: np.ndarray,
        ps_residuals: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Power spectrum frequency of the given time-series residuals.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals of an one-dimensional time-series.

        ps_residuals : :obj:`np.ndarray`, optional
            Power spectrum of ``ts_residuals``. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            Power spectrum frequency of the given time-series.

        References
        ----------
        .. [1] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time averaging
            over short, modified periodograms", IEEE Trans. Audio
            Electroacoust. vol. 15, pp. 70-73, 1967.
        .. [2] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        if ps_residuals is None:
            ps_residuals = cls._calc_ps_residuals(ts_residuals=ts_residuals)

        # Note: in the reference paper, it is used just the maximal value
        # of the power spectrum. However, to enable summarization, here
        # we return the whole power spectrum.
        return ps_residuals

    @classmethod
    def ft_ps_freqs(
        cls,
        ts_residuals: np.ndarray,
        freq_num: t.Union[int, float] = 0.05,
        ps_residuals: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Largest power spectrum frequencies of the given time-series.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals of an one-dimensional time-series.

        freq_num : int or float, optional
            If int >= 1, number of largest frequencies to be returned.
            If float in (0, 1), fraction of largest frequencies to be returned.

        ps_residuals : :obj:`np.ndarray`, optional
            Power spectrum of ``ts_residuals``. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            Largest power spectrum frequencies of the given time-series.

        References
        ----------
        .. [1] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time averaging
            over short, modified periodograms", IEEE Trans. Audio
            Electroacoust. vol. 15, pp. 70-73, 1967.
        .. [2] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        if freq_num <= 0:
            raise ValueError("'freq_num' must be positive.")

        if freq_num < 1:
            freq_num = np.ceil(freq_num * ts_residuals.size)

        freq_num = int(freq_num)

        if ps_residuals is None:
            ps_residuals = cls._calc_ps_residuals(ts_residuals=ts_residuals)

        ps_highest = np.sort(ps_residuals)[-freq_num:]

        return ps_highest

    @classmethod
    def ft_ps_peaks(
        cls,
        ts_residuals: np.ndarray,
        factor: float = 0.6,
        ps_residuals: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Number of significative power spectrum frequencies.

        Given a set of power spectrum frequencies `p`, a power spectrum
        frequency `f_i` is considered significative if and only if
        `f_i` >= factor * max(p), where `factor` is a user-defined
        parameter.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals of an one-dimensional time-series.

        factor : float, optional
            Multiplicative factor of the power spectrum maximum value to
            used to create the threshold to define which power spectrum
            frequencies are significative.

        ps_residuals : :obj:`np.ndarray`, optional
            Power spectrum of ``ts_residuals``. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            Binary array marking which time-series residuals power spectrum
            frequencies are larger than `factor * max(ps_residuals)`.

        References
        ----------
        .. [1] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time averaging
            over short, modified periodograms", IEEE Trans. Audio
            Electroacoust. vol. 15, pp. 70-73, 1967.
        .. [2] Lemke, Christiane & Gabrys, Bogdan. (2010). Meta-learning for
            time series forecasting and forecast combination. Neurocomputing.
            73. 2006-2016. 10.1016/j.neucom.2009.09.020.
        """
        if not 0 < factor < 1:
            raise ValueError("'factor' must be in (0, 1) range.")

        if ps_residuals is None:
            ps_residuals = cls._calc_ps_residuals(ts_residuals=ts_residuals)

        max_ps = np.max(ps_residuals)

        ps_peaks = (ps_residuals >= factor * max_ps).astype(int)

        # Note: in the reference paper, only the 'ps_peaks' sum is
        # returned. However, to enable summarization, here we return
        # the whole array.
        return ps_peaks

    @classmethod
    def ft_ps_entropy(
        cls,
        ts_residuals: np.ndarray,
        normalize: bool = True,
        base: int = 2,
        ps_residuals: t.Optional[np.ndarray] = None,
    ) -> float:
        """Spectral entropy of time-series residuals.

        The spectral entropy is the entropy if the normalized power spectrum
        of the detrended time-series. Technically, it is the entropy of the
        spectral density, which is the power spectrum normalized by the length
        of the time-series. However, this constant factor of normalization does
        not affect the entropy value.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals of an one-dimensional time-series.

        normalize : bool, optional (default=True)
            If True, the return value will be normalized in [0, 1].

        base : int, optional (default=2)
            Base of the entropy measure.

        ps_residuals : :obj:`np.ndarray`, optional
            Power spectrum of ``ts_residuals``. Used to take advantage of
            precomputations.

        Returns
        -------
        float
            Spectral entropy of time-series residuals.

        References
        ----------
        .. [1] P. Welch, "The use of the fast Fourier transform for the
            estimation of power spectra: A method based on time averaging
            over short, modified periodograms", IEEE Trans. Audio
            Electroacoust. vol. 15, pp. 70-73, 1967.
        .. [2] Pan, Y. N., J. Chen, and X. L. Li. "Spectral Entropy: A
            Complementary Index for Rolling Element Bearing Performance
            Degradation Assessment." Proceedings of the Institution of
            Mechanical Engineers, Part C: Journal of Mechanical Engineering
            Science. Vol. 223, Issue 5, 2009, pp. 1223–1231.
        .. [3] Shen, J., J. Hung, and L. Lee. "Robust Entropy-Based Endpoint
            Detection for Speech Recognition in Noisy Environments."
            ICSLP. Vol. 98, November 1998.
        .. [4] Vakkuri, A., A. Yli‐Hankala, P. Talja, S. Mustola, H.
            Tolvanen‐Laakso, T. Sampson, and H. Viertiö‐Oja. "Time‐Frequency
            Balanced Spectral Entropy as a Measure of Anesthetic Drug Effect
            in Central Nervous System during Sevoflurane, Propofol, and
            Thiopental Anesthesia." Acta Anaesthesiologica Scandinavica.
            Vol. 48, Number 2, 2004, pp. 145–153.
        """
        if ps_residuals is None:
            ps_residuals = cls._calc_ps_residuals(ts_residuals=ts_residuals)

        # Note: no need to calculate the power spectrum density 'd':
        # d = ps_residuals / ts_residuals.size
        # since a constant factor does not affect the entropy value.
        ps_ent = scipy.stats.entropy(
            ps_residuals / np.sum(ps_residuals), base=base
        )

        if normalize:
            ps_ent /= np.log(ts_residuals.size) / np.log(base)

        return ps_ent

    @classmethod
    def ft_low_freq_power(
        cls, ts_residuals: np.ndarray, threshold: float = 0.04
    ) -> float:
        """Proportion of low frequency in power spectrum using Hann window.

        Parameters
        ----------
        ts_residuals : :obj:`np.ndarray`
            Residuals of an one-dimensional time-series.

        threshold : float, optional (default=0.04)
            The threshold used to define how low is a `low frequency`.

        Returns
        -------
        float
            Proportion of low frequency in power spectrum of the time-series
            residuals, using a Hann window.

        References
        ----------
        .. [1] "Heart rate variability: Standards of measurement, physiological
            interpretation, and clinical use", M. Malik et al., Eur. Heart J.
            17(3) 354 (1996)
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        freqs, hann_ps = cls._calc_ps_residuals(
            ts_residuals=ts_residuals, window="hann", return_freqs=True
        )

        # Note: scale frequencies to the range [0, pi] (originally in [0, 0.5])
        # in order to be consistent with the Ben Fulcher code used as reference
        freqs *= 2 * np.pi

        # Note: no need to multiply both numerator and denominator by
        # freqs[1] - freqs[0], because these factor will cancel each other.
        low_freq_prop = np.sum(hann_ps[freqs <= threshold]) / np.sum(hann_ps)

        return low_freq_prop
